;-------------------------------------------------------------------------------
; Scrolltext en LowRes ou HiRes.
;
; Codé par Denis Duplan a.k.a. Yragael (stashofcode@gmail.com, stashofcode@gmail.com) en août 2018.
;
; Code & documentation on www.stashofcode.com (EN) and www.stashofcode.fr (FR)
;-------------------------------------------------------------------------------

;Cette oeuvre est mise à disposition selon les termes de la Licence (http://creativecommons.org/licenses/by-nc/4.0/) Creative Commons Attribution - Pas d’Utilisation Commerciale 4.0 International.

;Pour préciser si le scroll est en LowRes ou HiRes, définir la constante SCROLL_HIRES (0 : LowRes, 1 : HiRes).

;---------- Initialisation ----------

;Entrée(s) :
;	A0 = Adresse de la structure d'initialisation (sur le modèle de sctSetupData)
;Sortie(s) :
;	(rien)
;Notice :
;	Parce qu'il s'appuie sur le scrolling hardware, le scroll occupe toujours la largeur de l'écran. Les paramètres de largeur et de modulo s'interprètent ainsi :
;
;	- OFFSET_SCROLLSETUP_SCREENWIDTH : la largeur de la partie visible à l'écran
;
;	- OFFSET_SCROLLSETUP_SCREENMODULO : le # d'octets qu'il faut ajouter à une adresse après avoir parcouru OFFSET_SCROLLSETUP_SCREENWIDTH pour pointer sur le pixel de la ligne suivante dans le même bitplane
;
;	Attention! le modulo ne recoupe pas le modulo au sens de BPLxMOD. En effet, pour afficher un scroll de SCROLL_DX de large, il faut configurer un affichage dont la fenêtre (DIWSTRT/STOP) a une largeur de SCROLL_DX à partir de SCROLL_X. De plus, les données doivent être lues 16 pixels avant SCROLL_X, ce qui rajoute une colonne de 16 pixels jamais utilisée sur la gauche (cette colonne ne sert qu'à pouvoir afficher la second décalée de 0 : BPLxPTH/L pointent sur cette colonne et BPLCON1 vaut $0000). Enfin, il faut disposer de SCROLL_DX pixels de plus sur la droite pour le procédé d'affichage. Dans ces conditions, le bitplane doit avoir une largeur de 2*SCROLL_DX+16, et BPLxMOD doit donc valoir (SCROLL_DX-16)>>3. Toutefois, OFFSET_SCROLLSETUP_SCREENMODULO doit valoir (SCROLL_DX+16)>>3.

_sclSetup:
	movem.l d0-d1/a0-a1,-(sp)

	;Sauvegarder la structure d'initialisation

	lea sctState,a1
	move.w #DATASIZE_SCROLLSETUP-1,d0
_sclSetupCopySetup:
	move.b (a0)+,(a1)+
	dbf d0,_sclSetupCopySetup

	;Compléter l'initialisation

	lea sctState,a0
	move.l OFFSET_SCROLLSETUP_TEXT(a0),OFFSET_SCROLL_CHAR(a0)
	move.w #0,OFFSET_SCROLL_OFFSETX(a0)
	movea.l OFFSET_SCROLLSETUP_BPLCON1(a0),a1
	move.w #$0000,(a1)
	moveq #0,d0
	move.w OFFSET_SCROLLSETUP_SCREENWIDTH(a0),d0
	lsr.w #3,d0
	move.l OFFSET_SCROLLSETUP_BITPLANE(a0),d1
	add.l d0,d1
	movea.l OFFSET_SCROLLSETUP_BPLxPTHL(a0),a1
	move.w d1,4(a1)
	swap d1
	move.w d1,(a1)
	swap d1
	addq.l #1,d1
	move.l d1,OFFSET_SCROLL_LASTOUT(a0)
	add.l d0,d1
	move.l d1,OFFSET_SCROLL_LASTIN(a0)
	add.w OFFSET_SCROLLSETUP_SCREENMODULO(a0),d0
	moveq #0,d1
	move.b OFFSET_SCROLLSETUP_Y(a0),d1
	mulu d1,d0
	move.w d0,OFFSET_SCROLL_OFFSETY(a0)

	movem.l (sp)+,d0-d1/a0-a1
	rts

;---------- Itération ----------

;Entrée(s) :
;	(rien)
;Sortie(s) :
;	(rien)

_sclStep:
	movem.l d0-d2/a0-a4,-(sp)

	lea sctState,a0

	;Dessiner éventuellement des caractères pour l'étape courante. Un caractère est affiché tous les 8 pixels (la vitesse est contrainte à 1, 2, 4 ou 8 pixels, multiples de 8, si bien que l'événement peut être détecté en vérifiant si l'offset horizontal est multiple de 8)

	move.w OFFSET_SCROLL_OFFSETX(a0),d0
	and.b #$07,d0
	bne _sclStepNoPrint

	;Copier le dernier caractère affiché intégralement avant le premier caractère affiché intégralement

	move.w OFFSET_SCROLLSETUP_SCREENWIDTH(a0),d0
	lsr.w #3,d0
	add.w OFFSET_SCROLLSETUP_SCREENMODULO(a0),d0
	move.w OFFSET_SCROLL_OFFSETY(a0),d1
	movea.l OFFSET_SCROLL_LASTOUT(a0),a1
	movea.l OFFSET_SCROLL_LASTIN(a0),a2
	lea (a1,d1.w),a3
	lea (a2,d1.w),a4
	REPT 8
	move.b (a4),(a3)
	lea (a3,d0.w),a3
	lea (a4,d0.w),a4
	ENDR

	;Faire progresser les pointeurs sur le caractère précédent le premier caractère affiché intégralement, et sur le dernier caractère affiché intégralement, en tenant compte d'un possible rebouclage

	tst.w OFFSET_SCROLL_OFFSETX(a0)
	bne _sclStepInOutNoRestart
	lea 1(a1),a2
	move.l OFFSET_SCROLLSETUP_BITPLANE(a0),d2
	move.l OFFSET_SCROLLSETUP_BPLxPTHL(a0),a1
	move.w d2,4(a1)
	swap d2
	move.w d2,(a1)
	swap d2
	addq.l #2,d2
	move.l d2,a1
	bra _sclStepInOutDone
_sclStepInOutNoRestart:
	lea 1(a1),a1
	lea 1(a2),a2
_sclStepInOutDone:
	move.l a1,OFFSET_SCROLL_LASTOUT(a0)
	move.l a2,OFFSET_SCROLL_LASTIN(a0)

	;Dessiner le caractère suivant après le dernier caractère affiché intégralement

	lea (a2,d1.w),a2
	move.l OFFSET_SCROLL_CHAR(a0),a1
	moveq #0,d1
	move.b (a1)+,d1
	bne _sclStepNoTextRestart
	movea.l OFFSET_SCROLLSETUP_TEXT(a0),a1
	move.b (a1)+,d1
_sclStepNoTextRestart:
	move.l a1,OFFSET_SCROLL_CHAR(a0)
	subi.b #$20,d1
	lsl.w #3,d1
	movea.l OFFSET_SCROLLSETUP_FONT(a0),a1
	lea (a1,d1.w),a1
	REPT 8
	move.b (a1)+,(a2)
	lea (a2,d0.w),a2
	ENDR

_sclStepNoPrint:

	;Calculer l'offset pour l'étape suivante

	move.w OFFSET_SCROLL_OFFSETX(a0),d0
	moveq #0,d1
	move.b OFFSET_SCROLLSETUP_SPEED(a0),d1
	add.w d1,d0
	cmp.w OFFSET_SCROLLSETUP_SCREENWIDTH(a0),d0
	blt _sclStepOffsetXNoRestart
	moveq #0,d0
_sclStepOffsetXNoRestart:
	move.w d0,OFFSET_SCROLL_OFFSETX(a0)

	;Modifier BPLCON1 et BPLxPTH/L

	;LowRes: Les valeurs du scrolling hardware vont de 0 à 15 (chaque incrément décale le bitplane de 1 pixel LowRes)
	;
	; Offset | BPLCON1 | BPLxPTH/L
	;--------+---------+----------
	;    0   |  $0000  |    +0
	;    1   |  $00FF  |    +2
	;   ...
	;   15   |  $0011  |    +2
	;   16   |  $0000  |    +2
	;   17   |  $00FF  |    +4
	;   ...
	;   31   |  $0011  |    +4
	;   32   |  $0000  |    +4
	;   33   |  $00FF  |    +6
	;   ...
	;
	;=> Retard = (~(Offset-1))&$F
	;
	;HiRes: Les valeurs du scrolling hardware vont de 0 à 7 (chaque incrément décale le bitplane de 2 pixels HiRes)
	;
	; Offset | BPLCON1 | BPLxPTH/L
	;--------+---------+----------
	;    0   |  $0000  |    +0
	;    1   |  $0000  |    +0
	;    2   |  $0077  |    +2
	;    3   |  $0077  |    +2
	;   ...
	;   14   |  $0011  |    +2
	;   15   |  $0011  |    +2
	;   16   |  $0000  |    +2
	;   17   |  $0000  |    +2
	;   18   |  $0077  |    +4
	;   19   |  $0077  |    +4
	;   20   |  $0066  |    +4
	;   21   |  $0066  |    +4
	;   ...
	;   30   |  $0011  |    +4
	;   31   |  $0011  |    +4
	;   32   |  $0000  |    +4
	;   33   |  $0000  |    +4
	;   ...
	;
	;=> Retard = (~((Offset>>1)-1))&$7

	IFNE SCROLL_HIRES
	move.b d0,d1
	lsr.b #1,d1
	subq.b #1,d1
	not.b d1
	and.b #$07,d1
	move.b d1,d2
	lsl.b #4,d1
	or.b d2,d1	
	ELSE
	move.b d0,d1
	subq.b #1,d1
	not.b d1
	move.b d1,d2
	lsl.b #4,d1
	and.b #$0F,d2
	or.b d2,d1
	ENDC
	move.l OFFSET_SCROLLSETUP_BPLCON1(a0),a1
	move.w d1,(a1)

	IFNE SCROLL_HIRES
	cmpi.w #1,d0
	ble _sclStepNoBPLxPTHLUpdate
	ELSE
	tst.w d0
	beq _sclStepNoBPLxPTHLUpdate
	ENDC
	move.l OFFSET_SCROLLSETUP_BITPLANE(a0),a1
	IFNE SCROLL_HIRES
	subq.w #2,d0		;Bref, D0 = ((D0 - 2) / 16) * 2
	ELSE
	subq.w #1,d0		;Bref, D0 = ((D0 - 1) / 16) * 2
	ENDC
	lsr.w #3,d0
	and.b #$FE,d0
	addq.w #2,d0
	lea (a1,d0.w),a1
	move.l a1,d0
	move.l OFFSET_SCROLLSETUP_BPLxPTHL(a0),a1
	move.w d0,4(a1)
	swap d0
	move.w d0,(a1)
_sclStepNoBPLxPTHLUpdate:

	movem.l (sp)+,d0-d2/a0-a4
	rts

;---------- Finalisation ----------

;Entrée(s) :
;	(rien)
;Sortie(s) :
;	(rien)

_sclEnd:
	rts

;---------- Données ----------

sctSetupData:
OFFSET_SCROLLSETUP_BITPLANE=0
OFFSET_SCROLLSETUP_SCREENWIDTH=4
OFFSET_SCROLLSETUP_SCREENMODULO=6
OFFSET_SCROLLSETUP_BPLCON1=8
OFFSET_SCROLLSETUP_BPLxPTHL=12
OFFSET_SCROLLSETUP_FONT=16
OFFSET_SCROLLSETUP_TEXT=20
OFFSET_SCROLLSETUP_Y=24				;Ordonnée du texte en police 8 x 8 dans la bande du scroll
OFFSET_SCROLLSETUP_SPEED=25			;1, 2, 4 ou 8
DATASIZE_SCROLLSETUP=26
	BLK.B DATASIZE_SCROLLSETUP,0
sctState:
OFFSET_SCROLL_CHAR=DATASIZE_SCROLLSETUP
OFFSET_SCROLL_LASTIN=DATASIZE_SCROLLSETUP+4
OFFSET_SCROLL_LASTOUT=DATASIZE_SCROLLSETUP+8
OFFSET_SCROLL_OFFSETX=DATASIZE_SCROLLSETUP+12
OFFSET_SCROLL_OFFSETY=DATASIZE_SCROLLSETUP+14
DATASIZE_SCROLLSTATE=DATASIZE_SCROLLSETUP+16
	BLK.B DATASIZE_SCROLLSTATE,0

